require 'roundhouse'
require 'roundhouse/monitor'
require 'roundhouse/util'
require 'roundhouse/actor'

module Roundhouse
  ##
  # The Fetcher blocks on Redis, waiting for a message to process
  # from the queues.  It gets the message and hands it to the Manager
  # to assign to a ready Processor.
  class Fetcher
    include Util
    include Actor

    TIMEOUT = 1

    attr_reader :down

    def initialize(mgr, options)
      @down = nil
      @mgr = mgr
      @strategy = Fetcher.strategy.new(options)
    end

    # Fetching is straightforward: the Manager makes a fetch
    # request for each idle processor when Roundhouse starts and
    # then issues a new fetch request every time a Processor
    # finishes a message.
    #
    # Because we have to shut down cleanly, we can't block
    # forever and we can't loop forever.  Instead we reschedule
    # a new fetch if the current fetch turned up nothing.
    def fetch
      watchdog('Fetcher#fetch died') do
        logger.debug 'Fetcher terminating in #fetch' and return if Roundhouse::Fetcher.done?

        begin
          work = @strategy.retrieve_work
          ::Roundhouse.logger.info("Redis is online, #{Time.now - @down} sec downtime") if @down
          @down = nil

          if work
            @mgr.async.assign(work)
          else
            after(0) { fetch }
          end
        rescue => ex
          handle_fetch_exception(ex)
        end

      end
    end

    private

    def pause
      sleep(TIMEOUT)
    end

    def handle_fetch_exception(ex)
      if !@down
        logger.error("Error fetching message: #{ex}")
        ex.backtrace.each do |bt|
          logger.error(bt)
        end
      end
      @down ||= Time.now
      pause
      after(0) { fetch }
    rescue Celluloid::TaskTerminated
      # If redis is down when we try to shut down, all the fetch backlog
      # raises these errors.  Haven't been able to figure out what I'm doing wrong.
    end

    # Ugh.  Say hello to a bloody hack.
    # Can't find a clean way to get the fetcher to just stop processing
    # its mailbox when shutdown starts.
    def self.done!
      @done = true
      Roundhouse.logger.debug 'Fetcher: setting done'
    end

    def self.reset # testing only
      @done = nil
    end

    def self.done?
      defined?(@done) && @done
    end

    def self.strategy
      Roundhouse.options[:fetch] || RoundRobinFetch
    end
  end

  class RoundRobinFetch
    def initialize(options = nil)
      # ignore options
    end

    def retrieve_work
      work = Roundhouse.redis { |conn| Roundhouse::Monitor.maybe_next_job(conn) }
      UnitOfWork.new(*work) if work
    end

    # By leaving this as a class method, it can be pluggable and used by the Manager actor. Making it
    # an instance method will make it async to the Fetcher actor
    def self.bulk_requeue(inprogress, options)
      return if inprogress.empty?

      Roundhouse.logger.debug { "Re-queueing terminated jobs" }
      jobs_to_requeue = {}
      inprogress.each do |unit_of_work|
        jobs_to_requeue[unit_of_work.queue_id] ||= []
        jobs_to_requeue[unit_of_work.queue_id] << unit_of_work.message
      end

      Roundhouse.redis do |conn|
        conn.pipelined do
          jobs_to_requeue.each do |queue_id, jobs|
            Roundhouse::Monitor.requeue(conn, queue_id, jobs)
          end
        end
        # REFACTOR NOTE: This has to happen outside the pipelining since
        # we need to read. We can refactor to put this back
        # after converting the Monitor operations as EVAL scripts
        jobs_to_requeue.keys.each do |queue_id|
          Roundhouse::Monitor.push(conn, queue_id)
        end
      end
      Roundhouse.logger.info("Pushed #{inprogress.size} messages back to Redis")
    rescue => ex
      Roundhouse.logger.warn("Failed to requeue #{inprogress.size} jobs: #{ex.message}")
    end

    UnitOfWork = Struct.new(:full_queue_name, :message) do
      QUEUE_REGEX = /.*#{Roundhouse::Monitor::QUEUE}:/.freeze

      def acknowledge
        # nothing to do
      end

      def queue_id
        full_queue_name.gsub(QUEUE_REGEX, '')
      end

      def requeue
        Roundhouse.redis do |conn|
          Roundhouse::Monitor.requeue(conn, queue_id, message)
          Roundhouse::Monitor.push(conn, queue_id)
        end
      end
    end

  end
end
