require 'roundhouse/script'

module Roundhouse
  # This class implements two things:
  #   1. A turntable semaphore - the fetcher can pop the next available
  #      exclusive right to something (such as API request with a given
  #      auth token)
  #   2. Track which access right is temporarily suspended
  class Monitor
    ACTIVE    = 1
    EMPTY     = 0
    SUSPENDED = -1

    # This helps catch problems with key names at runtime
    TURNTABLE   = 'turntable'.freeze
    IN_ROTATION = 'inrotation'.freeze
    BUCKETS     = 'buckets'.freeze
    QUEUE       = 'queue'.freeze
    SCHEDULE    = 'schedule'.freeze
    STATUS      = 'status'.freeze

    # Number of seconds to block on turntable
    TURNTABLE_TIMEOUT = 5

    class << self
      # Find the first active queue
      # Return nil if nothing is there. The fetcher is responsible
      # for polling.
      def pop(conn)
        _, q_id = conn.brpop(TURNTABLE, TURNTABLE_TIMEOUT)
        conn.srem(IN_ROTATION, q_id)
        return q_id if queue_status(conn, q_id) == ACTIVE
        return nil
      end

      # Atomic push
      # Push queue into turntable if and only if queue status is active
      def push(conn, q_id)
        # NOTE: this version of redis-namespace has a bug when you do keys: argv: params
        TURNTABLE_PUSH.call conn, [status_bucket(q_id), TURNTABLE, IN_ROTATION], [q_id]
      end

      # Bulk requeue (push from right). Usually done
      # via Client, when Roundhouse is terminating
      def requeue(conn, q_id, jobs)
        conn.rpush("#{QUEUE}:#{q_id}", jobs)
      end

      def maybe_next_job(conn)
        queue_id = pop(conn)
        return nil unless queue_id

        job = pop_job(conn, queue_id)
        return queue_id, job if job
        return nil
      end

      # Atomically pop job
      # If nothing is in the queue, set queue status to empty
      def pop_job(conn, q_id)
        POP_JOB.call conn, ["#{QUEUE}:#{q_id}", status_bucket(q_id)], [q_id]
      end

      def push_job(conn, payloads)
        return schedule(conn, payloads) if payloads.first['at']

        q_id = payloads.first['queue_id']
        now = Time.now.to_f
        to_push = payloads.map do |entry|
          entry['enqueued_at'.freeze] = now
          Roundhouse.dump_json(entry)
        end
        conn.lpush("#{QUEUE}:#{q_id}", to_push)

        maybe_add_to_rotation(conn, q_id)
      end

      def set_queue_is_empty(conn, q_id)
        set_queue_status(conn, q_id, EMPTY)
      end

      def activate(conn, q_id)
        set_queue_status(conn, q_id, ACTIVE)
      end

      def suspend(conn, q_id)
        set_queue_status(conn, q_id, SUSPENDED)
      end

      def resume(conn, q_id)
        RESUME_QUEUE.call conn, [status_bucket(q_id), TURNTABLE, IN_ROTATION], [q_id]
      end

      def queue_status(conn, q_id)
        conn.hget(status_bucket(q_id), q_id).to_i || EMPTY
      end

      # Only push onto turntable is status is empty
      # If empty, push into turntable and set to active
      # Subsequent job pushes would see the active queue and
      # not push into turntable
      # Used after pushing a job, and conditionally putting it into rotation
      def maybe_add_to_rotation(conn, q_id)
        MAYBE_TURNTABLE_PUSH.call conn, [status_bucket(q_id), TURNTABLE, BUCKETS, IN_ROTATION], [q_id, bucket_num(q_id)]
      end

      def status_bucket(q_id)
        "#{STATUS}:#{bucket_num(q_id)}"
      end

      def bucket_num(q_id)
        q_id.to_i / 1000
      end

      # TODO: This needs to be in a lua script and atomic
      # Need to test how well this scales too. Should be testing about
      # 10,000 queues
      def rebuild_turntable!
        Roundhouse.redis do |conn|
          buckets = conn.smembers(BUCKETS)
          queues = conn.pipelined do
            buckets.each { |bucket| conn.hgetall("#{STATUS}:#{bucket}") }
          end

          all_queue_ids = queues.map(&:keys).flatten
          queue_len_res = conn.pipelined do
            all_queue_ids.each { |queue| conn.llen("#{QUEUE}:#{queue}") }
          end

          queue_len = all_queue_ids.zip(queue_len_res)
          status = queues.inject({}) { |a,x| a.merge(x) }

          conn.multi do
            conn.del(TURNTABLE)
            conn.del(IN_ROTATION)
            queue_len.each do |(q_id, len)|
              s = status[q_id].to_i

              case s
              when SUSPENDED then next
              when ACTIVE then
                if len > 0
                  conn.lpush(TURNTABLE, q_id)
                  conn.sadd(IN_ROTATION, q_id)
                else
                  set_queue_status(conn, q_id, EMPTY, false)
                end
              when EMPTY then
                next if len <= 0
                conn.lpush(TURNTABLE, q_id)
                conn.sadd(IN_ROTATION, q_id)
                set_queue_status(conn, q_id, ACTIVE, false)
              else
                set_queue_status(conn, q_id, SUSPENDED, false)
              end
            end
          end
        end
      end

      private

      def schedule(conn, payloads)
        conn.zadd(SCHEDULE.freeze, payloads.map do |hash|
              at = hash.delete('at'.freeze).to_s
              [at, Roundhouse.dump_json(hash)]
          end )
      end

      def set_queue_status(conn, q_id, status, add_bucket = true)
        conn.sadd(BUCKETS, bucket_num(q_id)) if add_bucket
        conn.hset(status_bucket(q_id), q_id, status)
      end
    end # class << self

    WOLVERINE_CONFIG = Roundhouse::Script::Configuration.new

    LUA_TURNTABLE_PUSH = <<END
local status = redis.call("HGET", KEYS[1], ARGV[1])
local in_rotation = redis.call("SISMEMBER", KEYS[3], ARGV[1])
if status == "#{ACTIVE}" and in_rotation == 0 then
  redis.call("SADD", KEYS[3], ARGV[1])
  return redis.call("LPUSH", KEYS[2], ARGV[1])
end
return nil
END
    LUA_MAYBE_TURNTABLE_PUSH = <<END
local status = redis.call("HGET", KEYS[1], ARGV[1])
if status == "#{EMPTY}" or status == nil or status == false then
  redis.call('SADD', KEYS[3], ARGV[2])
  redis.call("HSET", KEYS[1], ARGV[1], #{ACTIVE})
  local in_rotation = redis.call("SISMEMBER", KEYS[4], ARGV[1])
  if in_rotation == 0 then
    redis.call('SADD', KEYS[4], ARGV[1])
    return redis.call("LPUSH", KEYS[2], ARGV[1])
  end
end
return nil
END
    LUA_POP_JOB = <<END
local job = redis.call("RPOP", KEYS[1])
if type(job) == "string" then
  return job
end
redis.call("HSET", KEYS[2], ARGV[1], #{EMPTY})
return nil
END
    LUA_RESUME_QUEUE = <<END
local status = redis.call("HGET", KEYS[1], ARGV[1])
if status == "#{SUSPENDED}" then
  redis.call("HSET", KEYS[1], ARGV[1], #{ACTIVE})
  local in_rotation = redis.call("SISMEMBER", KEYS[3], ARGV[1])
  if in_rotation == 0 then
    redis.call('SADD', KEYS[4], ARGV[1])
    return redis.call("LPUSH", KEYS[2], ARGV[1])
  end
end
return nil
END
    TURNTABLE_PUSH       = Roundhouse::Script.new(LUA_TURNTABLE_PUSH.freeze, name: :turntable_push, config: WOLVERINE_CONFIG)
    MAYBE_TURNTABLE_PUSH = Roundhouse::Script.new(LUA_MAYBE_TURNTABLE_PUSH.freeze, name: :maybe_turntable_push, config: WOLVERINE_CONFIG)
    POP_JOB              = Roundhouse::Script.new(LUA_POP_JOB.freeze, name: :pop_job, config: WOLVERINE_CONFIG)
    RESUME_QUEUE         = Roundhouse::Script.new(LUA_RESUME_QUEUE.freeze, name: :resume_queue, config: WOLVERINE_CONFIG)

  end # Monitor
end
