require 'roundhouse/client'
require 'roundhouse/core_ext'

module Roundhouse

  ##
  # Include this module in your worker class and you can easily create
  # asynchronous jobs:
  #
  # class HardWorker
  #   include Roundhouse::Worker
  #
  #   def perform(*args)
  #     # do some work
  #   end
  # end
  #
  # Then in your Rails app, you can do this:
  #
  #   HardWorker.perform_async(queue_id, 1, 2, 3)
  #
  # Note that perform_async is a class method, perform is an instance method.
  module Worker
    attr_accessor :jid

    def self.included(base)
      raise ArgumentError, "You cannot include Roundhouse::Worker in an ActiveJob: #{base.name}" if base.ancestors.any? {|c| c.name == 'ActiveJob::Base' }

      base.extend(ClassMethods)
      base.class_attribute :roundhouse_options_hash
      base.class_attribute :roundhouse_retry_in_block
      base.class_attribute :roundhouse_retries_exhausted_block
    end

    def logger
      Roundhouse.logger
    end

    module ClassMethods

      def perform_async(queue_id, *args)
        client_push('queue_id' => queue_id, 'class' => self, 'args' => args)
      end

      def perform_in(queue_id, interval, *args)
        int = interval.to_f
        now = Time.now
        ts = (int < 1_000_000_000 ? (now + interval).to_f : int)

        item = { 'queue_id' => queue_id, 'class' => self, 'args' => args, 'at' => ts }

        # Optimization to enqueue something now that is scheduled to go out now or in the past
        item.delete('at'.freeze) if ts <= now.to_f

        client_push(item)
      end
      alias_method :perform_at, :perform_in

      ##
      # Allows customization for this type of Worker.
      # Legal options:
      #
      #   :retry - enable the RetryJobs middleware for this Worker, default *true*
      #   :backtrace - whether to save any error backtrace in the retry payload to display in web UI,
      #      can be true, false or an integer number of lines to save, default *false*
      #   :pool - use the given Redis connection pool to push this type of job to a given shard.
      def roundhouse_options(opts={})
        self.roundhouse_options_hash = get_roundhouse_options.merge(opts.stringify_keys)
      end

      def roundhouse_retry_in(&block)
        self.roundhouse_retry_in_block = block
      end

      def roundhouse_retries_exhausted(&block)
        self.roundhouse_retries_exhausted_block = block
      end

      def get_roundhouse_options # :nodoc:
        self.roundhouse_options_hash ||= Roundhouse.default_worker_options
      end

      def client_push(item) # :nodoc:
        pool = Thread.current[:roundhouse_via_pool] || get_roundhouse_options['pool'] || Roundhouse.redis_pool
        Roundhouse::Client.new(pool).push(item.stringify_keys)
      end

    end
  end
end
